//----------------------------------------------------------------------------
//
// Module: unit_image.pas (Part of Photometry Express)
// Author: J C Moore
//
// Copyright (C) 2013  J C Moore
//
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation, either version 3 of
// the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty
// of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
// See the GNU General Public License for more details.
//
// see http://www.gnu.org/licenses/
//
// Update History:
//    05 Mar 2013: last update by J C Moore
//    17 Nov 2013: Code commented by P Campbell-Burns
//
//  Notes:
//    FITs file uses array of big endian signed 16 bit integers
//
//----------------------------------------------------------------------------

unit Unit_Image;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, Math, StrUtils, Unit_Main, Unit_WCS;

const
  MAXRECORDS	= 36;		// for FITs files
  DPR         = 57.295779;
  {$I 'Photo_Express_Errors.inc'}

type

  TImageArray = Array of Array of Word;  // Fits bitmap

  TForm_Image = class(TForm)
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormPaint(Sender: TObject);

    procedure Prepare_Image;
    procedure Fill_Bitmap;
    procedure FormCreate(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure FormMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure FormMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormDeactivate(Sender: TObject);

  private
    Image_Array       : TImageArray;		// unsigned integers
    Image_Bmp         : TBitmap;		// to be displayed
    Histogram		      : Array[0..65535] of Cardinal;
    StartX,StartY     : Integer;
    Painting		      : Boolean;
    Start_BlackPoint	: Word;
    Start_WhitePoint	: Word;
    BlackPoint		    : Word;
    WhitePoint		    : Word;
    BitPix		        : SmallInt;
    KeyNameString	    : String;
    ValueString		    : String;
    CommentString	    : String;
    LineBuff		      : Array of Byte;
    Caption_String	  : String;
    Npap_real         : Real;           // number of pixels in star aperture
    Npan              : Integer;        // number of pixels in the annulus
    Net_Flux          : Real;
    Sky_Fluxpp        : Real;           // mean sky flux per pixel
    Ref_Background    : Real;           // ADUs per pixel
    Star_Background   : Real;           // ADUs per pixel

    // Plate solve coefficients:
    CD1_1_found   : Boolean;            // FITS scale and rotate matrix
    CD1_2_found   : Boolean;            // FITS scale and rotate matrix
    CD2_1_found   : Boolean;            // FITS scale and rotate matrix
    CD2_2_found   : Boolean;            // FITS scale and rotate matrix
    CRPIX1_found  : Boolean;            // Reference pixel on the horizonal axis
    CRPIX2_found  : Boolean;            // Reference pixel on the vertical axis
    CRVAL1_found  : Boolean;            // Horiz axis WCS value at the reference pixel
    CRVAL2_found  : Boolean;            // Vert axis WCS value at the reference pixel

    procedure Show_Image_Array(Which_CCD: Word);
    procedure Invert_Matrix();
    procedure CMMouseEnter(var msg: TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var msg: TMessage); message CM_MOUSELEAVE;
    procedure StopFlicker(var Msg: TWMEraseBkgnd); message WM_ERASEBKGND;

    function Measure_Flux(X,Y: Double; DoSnap: Boolean): Boolean;

  public
    FileName		      : String;
    KeyNames		      : Array[0..4*MAXRECORDS] of String;
    Values		        : Array[0..4*MAXRECORDS] of String;
    CommentStrings    : Array[0..4*MAXRECORDS] of String;
    NKeys		          : Word;
    XSize		          : Word;		// full or subframe
    YSize		          : Word;		// ditto
    Current_Scale	    : Integer;
    Plate_Solved      : Boolean;
    WCS_Calculator    : TWCS;
    WCS1              : TWCSRecord;
    ObjectX,ObjectY   : Double;
    NewX,NewY         : Double;
    ObjectName        : String;
    Object_Marked     : Boolean;
    ReferenceX,ReferenceY: Double;  // 0,0 = top left of bitmap
    ReferenceName     : String;
    Reference_Marked  : Boolean;
    Last_Error        : Integer;
    Magnitude         : Double;
    Reference_Mag     : Double;
    Std_Error         : Real;       // magnitudes
    Less_Than         : Integer;
    CMag              : Real;       // instrumental magnitude of ref star
    JD                : Real;       // middle of exposure
    AirMass           : Real;
    Exposure          : Real;
    Date_Obs          : String;
    Image_Filter_Letter: Char;
    Ref_Filter_Letter : Char;

    Outburst          : Boolean;    // set by Main Form
    Threshold         : Double;     // ditto
    ChartName         : String;     // ditto
    AAVSO_Written     : Boolean;    // ditto

    procedure Paint_Object();
    procedure Paint_Reference();
    procedure ReScale(Scale: Integer);
    procedure Open_File(FName: String);
    procedure Reset_Visualisation();
    procedure Mark_Object(pos: TSkyPosition; name: String);
    procedure Mark_Reference(pos: TSkyPosition; name: String; mag: Double; Filter_Letter: char);

    function Measure_Magnitude(): Boolean;
  end;

var
  Form_Image: TForm_Image;

implementation

{$R *.DFM}

procedure TForm_Image.StopFlicker(var Msg: TWMEraseBkgnd);
begin
  Msg.Result	:= 1;
end;

procedure TForm_Image.Invert_Matrix();
var
  Denom : Double;
begin
  Denom       := WCS1.cd[3]*WCS1.cd[0] - WCS1.cd[1]*WCS1.cd[2];
  WCS1.dc[0]  := WCS1.cd[3]/Denom;
  WCS1.dc[1]  := -WCS1.cd[1]/Denom;
  WCS1.dc[2]  := -WCS1.cd[2]/Denom;
  WCS1.dc[3]  := WCS1.cd[0]/Denom;
end;

procedure TForm_Image.Reset_Visualisation();
begin
  Prepare_Image();
  Refresh;
end;

// Import a FITs file:
//
procedure TForm_Image.Open_File(FName: String);
var
  F               : File;
  FITsRecord	    : Array[1..80] of Byte;	// one FITs record
  Done		        : Boolean;
  BlockCounter	  : Integer;
  i,j,k		        : Integer;
  c		            : AnsiChar;
  wc              : Char;
  Found_Comment	  : Boolean;
  Found_String	  : Boolean;
  Found_Value_Start: Boolean;
  BScale	        : Extended;
  BZero		        : Word;
  pixel_cardinal  : Cardinal;
//  NAxis3	      : Word;

begin
  WCS_Calculator  := TWCS.Create;
  // Default values:
  //
  WCS1.coorflip := False;
  WCS1.longpole := 180.0;
//  WCS1.latpole  := 0.0;
  WCS1.lngcor   := nil;
  WCS1.latcor   := nil;
  WCS1.rodeg    := DPR;

  Caption_String	:= FName;
  FileName		    := FName;
  AssignFile(F,FName);
  FileMode		    := 2;		                // read and write
  Reset(F,1);				                      // record size = 1

// Read in 2880 byte header(s):
//
  Done		:= False;
  BlockCounter	:= 0;
  CD1_1_found := False;
  CD1_2_found := False;
  CD2_1_found := False;
  CD2_2_found := False;
  Repeat
    for j:=1 to MAXRECORDS do
    begin
      BlockRead(F, FITsRecord, 80);		// read a record at a time
      KeyNameString	:= '';
      for i:=1 to 8 do                // Extract first 8 chars, build FITS keyword
      begin
        c             := AnsiChar(FITsrecord[i]);
        wc            := Char(c);
        KeyNameString := KeyNameString + wc;
      end;
      KeyNameString	:= TrimRight(KeyNameString);
      KeyNames[BlockCounter]	:= KeyNameString;
      if KeyNameString = 'END' then
      begin
        Done	:= True;
	      NKeys	:= BlockCounter			// excluding the 'END'
      end;
      ValueString   := '';
      CommentString	:= '';
      Found_Comment	:= False;
      Found_String	:= False;
      Found_Value_Start	:= False;
      if FITsRecord[9] = ord('=') then
      begin					// value found
        if FITsRecord[10] <> ord(' ') then
          ShowMessage('"= " Expected!')
        else
	      begin
          for i := 10 to 80 do
          begin
            c   := AnsiChar(FITsRecord[i]);
            wc  := Char(c);
            if Found_String then
            begin
	            ValueString := ValueString + wc;
              if c = '''' then
                Found_String	:= False
            end else
            begin
              if Found_Comment then
                CommentString := CommentString + wc
              else
	            begin			// neither comment nor string
                if c = '/' then
                  Found_Comment := True
                else
		            begin
                  if Found_Value_Start then
                    ValueString := ValueString + wc
                  else if c <> ' ' then
		              begin
                    Found_Value_Start := True;
                    ValueString := ValueString + wc;
                  end;			// c <> ' '
                  if c = '''' then
                    Found_String := True
                end			// found comment <> true
              end			// not found comment
            end				// not found string
	        end				// for i:= 1 to 70 do
        end				// "= " Expected
      end				// value found
      else begin                        // value not found, treat as comment
        for i:=10 to 80 do begin
          c   := AnsiChar(FITsRecord[i]);
          wc  := Char(c);
          CommentString := CommentString + wc;
        end;
      end;
      Values[BlockCounter] := TrimRight(ValueString);
      CommentStrings[BlockCounter] := TrimLeft(CommentString);

      if KeyNameString = 'EXPOSURE' then
      begin
        try
          Exposure  := StrtoFloat(ValueString);
        except
          Exposure  := 0.0;
        end;    
      end;  

      if KeyNameString = 'JD' then
      begin
        try
          JD  := StrtoFloat(ValueString);
        except
          JD  := 0.0;
        end;    
      end;  

      if KeyNameString = 'AIRMASS' then
      begin
        try
          AirMass  := StrtoFloat(ValueString);
        except
          AirMass  := 1.0;
        end;    
      end;  

      if KeyNameString = 'DATE-OBS' then
      begin
        Date_Obs  := ValueString;
      end;  

      // Collect plate solve coefficients, if present:
      //
      if KeyNameString = 'CD1_1' then
      begin
        try
          WCS1.cd[0]  := StrtoFloat(ValueString);
          CD1_1_found := True;
        except
          WCS1.cd[0]  := 0.0;
        end;    
      end;  

      if KeyNameString = 'CD1_2' then
      begin
        try
          WCS1.cd[1]  := StrtoFloat(ValueString);
          CD1_2_found := True;
        except
          WCS1.cd[1]  := 0.0;
        end;    
      end;  

      if KeyNameString = 'CD2_1' then
      begin
        try
          WCS1.cd[2]  := StrtoFloat(ValueString);
          CD2_1_found := True;
        except
          WCS1.cd[2]  := 0.0;
        end;    
      end;  

      if KeyNameString = 'CD2_2' then
      begin
        try
          WCS1.cd[3]  := StrtoFloat(ValueString);
          CD2_2_found := True;
        except
          WCS1.cd[3]  := 0.0;
        end;    
      end;  

      if KeyNameString = 'CRPIX1' then
      begin
        try
          WCS1.crpix[0] := StrtoFloat(ValueString);
          WCS1.xrefpix  := StrtoFloat(ValueString);
          CRPIX1_found  := True;
        except
          WCS1.crpix[0] := 0.0;
          WCS1.xrefpix  := 0.0;
        end;    
      end;  

      if KeyNameString = 'CRPIX2' then
      begin
        try
          WCS1.crpix[1] := StrtoFloat(ValueString);
          WCS1.yrefpix  := StrtoFloat(ValueString);
          CRPIX2_found  := True;
        except
          WCS1.crpix[1] := 0.0;
          WCS1.yrefpix  := 0.0;
        end;    
      end;  

      if KeyNameString = 'CRVAL1' then
      begin
        try
          WCS1.crval[0]:= StrtoFloat(ValueString);
          CRVAL1_found := True;
        except
          WCS1.crval[0]:= 0.0;
        end;    
      end;  

      if KeyNameString = 'CRVAL2' then
      begin
        try
          WCS1.crval[1]:= StrtoFloat(ValueString);
          CRVAL2_found := True;
        except
          WCS1.crval[1]:= 0.0;
        end;    
      end;  

      if KeyNameString = 'CDELT1' then
      begin
        try
          WCS1.cdelt[0]:= StrtoFloat(ValueString);
        except
          WCS1.cdelt[0]:= 0.0;
        end;    
      end;  

      if KeyNameString = 'CDELT2' then
      begin
        try
          WCS1.cdelt[1]:= StrtoFloat(ValueString);
        except
          WCS1.cdelt[1]:= 0.0;
        end;    
      end;  

      if KeyNameString = 'LONPOLE' then
      begin
        try
          WCS1.longpole:= StrtoFloat(ValueString);
        except
          WCS1.longpole:= 180.0;
        end;    
      end;  

      if KeyNameString = 'FILTER' then
        Image_Filter_Letter := ValueString[2];

      Inc(BlockCounter,1);
    end					// for j
  Until Done;

  if CD1_1_found and CD1_2_found and CD2_1_found and CD2_2_found
    and CRPIX1_found and CRPIX2_found and CRVAL1_found and CRVAL2_found then
  begin  
    Plate_Solved  := True;
    WCS1.rotmat   := True;
    Invert_Matrix();
  end else begin  
    Last_Error    := ERR_NOT_PLATE_SOLVED;
    Plate_Solved  := False;
    WCS1.rotmat   := False;
  end;  

  if KeyNames[0] = 'SIMPLE' then
  begin
    if Values[0] <> 'T' then
      ShowMessage('Not a simple FITs Format!')
  end else
    ShowMessage('SIMPLE keyword not found!');

  if KeyNames[1] = 'BITPIX' then
    BitPix	:= StrtoInt(Values[1])
  else  
    ShowMessage('BITPIX keyword not found!');
  if BitPix <> 16 then
    ShowMessage('Not a 16 bit image!');  

  if KeyNames[2] = 'NAXIS' then
  begin
    if StrtoInt(Values[2]) <> 2 then
      ShowMessage('Not a 2 dimensional image!')
  end else  
    ShowMessage('NAXIS keyword not found!');

  XSize := 0;
  for i:=3 to NKeys do
  begin
    if KeyNames[i] = 'NAXIS1' then
    begin
      XSize := StrtoInt(Values[i]);
      break
    end  
  end;

  YSize := 0;
  for i:=3 to NKeys do
  begin
    if KeyNames[i] = 'NAXIS2' then
    begin
      YSize := StrtoInt(Values[i]);
      break
    end  
  end;

  BZero	:= 0;
  for i:=3 to NKeys do
  begin
    if KeyNames[i] = 'BZERO' then
    begin
      BZero := trunc(StrtoFloat(Values[i])+0.5);
      break
    end
  end;

  BScale	:= 1.0;
  for i:=3 to NKeys do
  begin
    if KeyNames[i] = 'BSCALE' then
    begin
      BScale := StrtoFloat(Values[i]);
      break
    end
  end;  
  if BScale <> 1.0 then
    ShowMessage('BScale <> 1.0 not supported!');

  // Now read in and store the image data:
  //
  SetLength(Image_Array, YSize, XSize);
  SetLength(LineBuff, 2*XSize);

  for i:= 0 to YSize-1 do
  begin
    BlockRead(F, LineBuff[0], 2*XSize);	// read a line at a time MSByte first
    k   := 0;
    for j:= 0 to XSize-1 do
    begin

// Turn big endian integer into little endian unsigned:
//
      pixel_cardinal	:= LineBuff[k]*256 + LineBuff[k+1] + BZero;
      Image_Array[i,j]:= Word(pixel_cardinal);
      Inc(k,2);
    end;  
  end;
  CloseFile(F);
  JD  := JD + Exposure/172800;    // mid exposure
  Show_Image_Array(0);
end;

procedure TForm_Image.ReScale(Scale: Integer);
begin
 if Scale <> Current_Scale then
  begin
    Painting		:= False;
    Current_Scale	:= Scale;
    Main_Form.ComboBox_Scale.ItemIndex := Scale;	// necessary
    if Scale < 7 then begin
      ClientHeight	:= YSize div (8 - Scale);
      ClientWidth	:= XSize div (8 - Scale);
    end else begin
      ClientHeight	:= YSize * (Scale - 6);
      ClientWidth	:= XSize * (Scale - 6);
    end;
    Painting		:= True;
    FormPaint(Self);
  end;
end;

procedure TForm_Image.Fill_Bitmap();
var
  Range	: Word;
  x,y		: Word;
  bit		: Byte;
  x3		: Word;
  P     : PByteArray;
  px		: Real;
begin
  if WhitePoint > BlackPoint then
    Range	:= WhitePoint - BlackPoint
  else
    Range	:= 10;
  for y:=0 to YSize -1 do
  begin
    P	:= Image_Bmp.ScanLine[y];	// points to start of next line
    x3	:= 0;
    for x:=0 to XSize -1 do
    begin
      px	:= 256*(Image_Array[y,x] - BlackPoint)/Range;
      if px < 0 then
        px	:= 0.0
      else if px > 255.0 then
        px	:= 255.0;
      bit	:= trunc(px + 0.5);
      P[x3]	:= bit;			// blue
      P[x3+1]	:= bit;			// green
      P[x3+2]	:= bit;			// red
      Inc(x3,3);
    end;
  end;
  if Object_Marked then
    Paint_Object();
  if Reference_Marked then
    Paint_Reference();
end;

procedure TForm_Image.Prepare_Image();
var
  Total		    : Cardinal;
  Threshold	  : Cardinal;
  i		        : Word;
  BP,WP,Mean  : Word;
  NPixels     : Cardinal;
begin
  BP		:= 0;
  WP		:= 65535;
  NPixels	:= XSize*YSize;
  Total		:= 0;
  Threshold	:= NPixels div 50;
  for i:=0 to 65535 do
  begin
    Total := Total + Histogram[i];
    if Total > Threshold then
    begin
      BP	:= i;
      break;
    end;  
  end;
  Total		:= 0;
  Threshold	:= NPixels div 400;
  for i:= 65535 downto 0 do
  begin
    Total := Total + Histogram[i];
    if Total > Threshold then
    begin
      WP	:= i;
      break;
    end;  
  end;
  if BP >= WP then
  begin
    Mean := (BP+WP) div 2;
    if Mean < 100 then
      BP := 0 else
    BP := Mean -100;
    if Mean > 65435 then
      WP := 65535 else
    WP := Mean + 100;  
  end;
  BlackPoint	:= BP;
  WhitePoint	:= WP;
  Main_Form.Panel_BP.Caption	:= 'BP=' + InttoStr(BlackPoint);
  Main_Form.Panel_WP.Caption	:= 'WP=' + InttoStr(WhitePoint);
  Fill_Bitmap();
end;

procedure TForm_Image.Show_Image_Array(which_CCD: Word);
const
  RESOLUTION	= 1024;
var
  x,y		: Word;
  pixel		: Word;
  New_Buffer	: Boolean;
begin
  New_Buffer	:= False;
  if Image_Bmp = nil then begin
    New_Buffer			      := True;
    Image_Bmp			        := TBitmap.Create;
    Image_Bmp.HandleType	:= bmDIB;
    Image_Bmp.PixelFormat := pf24bit;
    Image_Bmp.Height		  := YSize;
    Image_Bmp.Width		    := XSize;
  end;  

  for y:= 0 to 65535 do
     Histogram[y] := 0;
  for y:= 0 to YSize -1 do
    for x:=0 to XSize -1 do
    begin
      pixel := Image_Array[y,x];
      Histogram[pixel] := Histogram[pixel] + 1;
    end;

  Prepare_Image();

  if New_Buffer then begin
    WindowState		:= wsNormal;
    Position		:= poDefault;
    Current_Scale	:= -1;		// Force a rescale
    ReScale(Main_Form.Scale_Index);	// paint new image and set Current_Scale
  end;  
  Refresh;
end;

procedure TForm_Image.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  if not(Image_Bmp=nil) then
    Image_Bmp.Free;
  if not(WCS_Calculator=nil) then
    WCS_Calculator.Free;
  if not(Image_Array=nil) then
    Image_Array		:= nil;
  if Main_Form.MDIChildCount < 2 then
  begin  
    Main_Form.Hide_Zoom();
  end;  
  Main_Form.Clear_Top_Panels;
  Action		:= caFree;
  Main_Form.Trigger_Results();
end;

procedure TForm_Image.FormPaint(Sender: TObject);
begin
  if Painting then
  begin
    Caption	:= Caption_String;
    Canvas.StretchDraw(ClientRect,Image_Bmp);	// using bitblt
  end;
end;

procedure TForm_Image.FormMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  StartX := X;
  StartY := Y;
  if ssShift in Shift then	// Shift key is pressed
  begin
    Cursor		:= crSize;
    Start_BlackPoint	:= BlackPoint;
    Start_WhitePoint	:= WhitePoint;
  end  
end;

procedure TForm_Image.FormMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
var
  deltaX,deltaY 	  : Integer;
  pixel			        : Word;
  ScaledX,ScaledY   : Integer;
  BB,WW,Mean,Range  : Double;
  Temp              : TSkyPosition;
  RA,Dec            : Double;
  RAMin,RASS        : Double;
  RAHH,RAMM         : Integer;
  DecMin,DecSS      : Double;
  DecHH,DecMM       : Integer;
  PosDec            : Double;
  Minus             : Boolean;
begin
  if (ssShift in Shift) and (ssLeft in Shift) then
// Change the visualisation:
  begin
    deltaX	:= X - StartX;
    deltaY	:= Y - StartY;
    if ((deltaX mod 5) = 0) or ((deltaY mod 5) = 0) then 
    begin
      BB		:= Start_BlackPoint;		// start of move
      WW		:= Start_WhitePoint;		// start if move
      if deltaY > 0 then
        Mean	:= (BB + WW)*(200.0 + deltaY)/200.0
      else
        Mean	:= (BB + WW)*200.0/(200.0 - deltaY);
      if deltaX > 0 then
        Range	:= (WW - BB)*200.0/(200.0 + deltaX)
      else
        Range	:= (WW - BB)*(200.0 - deltaX)/200.0;
  
      BB		:= trunc(0.5 + (Mean-Range)/2.0);
      WW		:= trunc(0.5 + (Mean+Range)/2.0);
      if (BB < 10.0) then BB:= 10.0;
      if (WW> 65000) then WW:= 65000;
      BlackPoint	:= trunc(BB);
      WhitePoint	:= trunc(WW);
      Main_Form.Panel_BP.Caption	:= 'BP=' + InttoStr(BlackPoint);
      Main_Form.Panel_WP.Caption	:= 'WP=' + InttoStr(WhitePoint);
      Fill_Bitmap();			// using the new BP & WP
      FormPaint(Self);
    end
  end else if (ssCtrl in Shift) and (ssLeft in Shift) then
  begin
    with Main_Form.VertScrollBar do
    begin
      Range := Main_Form.ClientHeight;
      Position := 700;
      ShowMessage('Control');
    end;
  end else				// Shift key is not pressed
                  // Show pixel location & value:
  begin
    if Current_Scale < 7 then begin
      ScaledX		:= X * (8 - Current_Scale);
      ScaledY		:= Y * (8 - Current_Scale);
    end else begin
      ScaledX		:= X div (Current_Scale - 6);
      ScaledY		:= Y div (Current_Scale - 6);
    end;
    if ScaledX < 0 then
      ScaledX := 0;
    if ScaledX > XSize-1 then
      ScaledX := XSize-1;
    if ScaledY < 0 then
      ScaledY := 0;
    if ScaledY > YSize-1 then
      ScaledY := YSize-1;
    pixel := Image_Array[ScaledY,ScaledX];
    Main_Form.Panel_Location.Caption := '(' + InttoStr(ScaledX) + ','
    + InttoStr(ScaledY) + ')';
    Main_Form.Panel_Pixel.Caption := 'I=' + InttoStr(pixel);

    if Plate_Solved then
    begin
      Temp  := WCS_Calculator.tnxpos(ScaledX,ScaledY,@WCS1);
      RA    := Temp.RA/15.0;
      Dec   := Temp.Dec;
      RAHH  := floor(RA);
      RAMin := 60.0*(RA - RAHH);
      RAMM  := floor(RAMin);
      RASS  := 60.0*(RAMin - RAMM);

      if Dec < 0.0 then
      begin  
        Minus   := True;
        PosDec  := -Dec
      end else
      begin
        Minus   := False;
        PosDec  := Dec;
      end;  
      DecHH := trunc(PosDec);
      DecMin:= 60.0*(PosDec - DecHH);
      DecMM := floor(DecMin);
      DecSS := 60.0*(DecMin - DecMM);

      if Minus then
        Main_Form.Panel_Position.Caption  := format('RA = %2d:%2d:%2.2f, Dec = -%d:%d:%2.2f',[RAHH,RAMM,RASS,DECHH,DecMM,DecSS])
      else  
        Main_Form.Panel_Position.Caption  := format('RA = %2d:%2d:%2.2f, Dec = %d:%d:%2.2f',[RAHH,RAMM,RASS,DECHH,DecMM,DecSS]);
    end;
  end;
end;

procedure TForm_Image.FormMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if ssShift in Shift then		// Shift key is pressed
    Refresh;
  Cursor := crDefault;
end;

procedure TForm_Image.FormCreate(Sender: TObject);
begin
  Image_Array           := nil;
  BorderIcons           := [biSystemMenu,biMinimize];
  BorderStyle	          := bsSingle;
  WindowState	          := wsMinimized;
  Constraints.MaxWidth  := 32767;
  Constraints.MaxHeight := 32767;
  Painting	            := True;
  Object_Marked         := False;
  Reference_Marked      := False;
  Last_Error            := ERR_NO_ACTIVE_IMAGE;
  AAVSO_Written         := False;
  Outburst              := False;
  ChartName             := '';
  Reference_Mag         := 0.0;
  CMag                  := 0.0;
  Exposure              := 0.0;
  Std_Error             := 0.0;
  Less_Than             := 0;
  AirMass               := 1.0;
end;

procedure TForm_Image.FormActivate(Sender: TObject);
begin
  Main_Form.ComboBox_Scale.ItemIndex := Current_Scale;
  Main_Form.Panel_BP.Caption	:= 'BP=' + InttoStr(BlackPoint);
  Main_Form.Panel_WP.Caption	:= 'WP=' + InttoStr(WhitePoint);
  Main_Form.Show_Results();
end;

procedure TForm_Image.FormResize(Sender: TObject);
begin
  FormPaint(Self);
end;

procedure TForm_Image.Paint_Reference();
var
  radius1   : Integer;
  radius2   : Integer;
  radius3   : Integer;
  intx,inty : Integer;
begin
  if Reference_Marked then
  begin  
    intx    := round(ReferenceX);
    inty    := round(ReferenceY);
    radius1 := Main_Form.Radius;
    radius2 := radius1 + Main_Form.Gap;
    radius3 := radius2 + Main_Form.Annulus;
    with Image_Bmp.Canvas do
    begin
      Pen.Color		:= clWhite;
      Pen.Style		:= psSolid;
      Pen.Width		:= 1;
      Brush.Style	:= bsClear;
      Ellipse(intx-radius1,inty-radius1,intx+radius1,inty+radius1);
      if Main_Form.Show_Annulus then
      begin  
        Ellipse(intx-radius2,inty-radius2,intx+radius2,inty+radius2);
        Ellipse(intx-radius3,inty-radius3,intx+radius3,inty+radius3);
      end;  
      Font.Name   := 'Aerial';
      Font.Size   := 24;
      Font.Color  := clYellow;
      Textout(intx+20,inty-20,ReferenceName);
    end;
    Refresh;
  end;
end;

procedure TForm_Image.Paint_Object();
var
  radius1   : Integer;
  radius2   : Integer;
  radius3   : Integer;
  intx,inty : Integer;
begin
  if Object_Marked then
  begin
    intx    := round(ObjectX);
    inty    := round(ObjectY);
    radius1 := Main_Form.Radius;
    radius2 := radius1 + Main_Form.Gap;
    radius3 := radius2 + Main_Form.Annulus;
    with Image_Bmp.Canvas do
    begin
      Pen.Color		:= clWhite;
      Pen.Style		:= psSolid;
      Pen.Width		:= 1;
      Brush.Style	:= bsClear;
      Ellipse(intx-radius1,inty-radius1,intx+radius1,inty+radius1);
      if Main_Form.Show_Annulus then
      begin  
        Ellipse(intx-radius2,inty-radius2,intx+radius2,inty+radius2);
        Ellipse(intx-radius3,inty-radius3,intx+radius3,inty+radius3);
      end;  
      Font.Name   := 'Aerial';
      Font.Size   := 24;
      Font.Color  := clGreen;
      Textout(intx+20,inty-20,ObjectName);
    end;  
    Refresh;
  end;  
end;  

procedure TForm_Image.Mark_Reference(pos: TSkyPosition; name: String; mag: Double; Filter_Letter: char);
var
  PPos    : TCCDPosition;
begin
  if Plate_Solved then
  begin  
    PPos          := WCS_Calculator.tnxpix(pos.RA*15.0,pos.Dec,@WCS1);
    ReferenceX    := PPos.x;
    ReferenceY    := PPos.y;
    ReferenceName := name;
    Reference_Marked := True;
    Reference_Mag := mag;
    Ref_Filter_Letter := Filter_Letter;
  end;  
end;

procedure TForm_Image.Mark_Object(pos: TSkyPosition; name: String);
var
  PPos    : TCCDPosition;
begin
  if Plate_Solved then
  begin  
    PPos          := WCS_Calculator.tnxpix(pos.RA*15.0,pos.Dec,@WCS1);
    ObjectX       := PPos.x;
    ObjectY       := PPos.y;
    ObjectName    := name;
    Object_Marked := True;
  end;  
end;

procedure TForm_Image.FormDeactivate(Sender: TObject);
begin
  Main_Form.Clear_Bottom_Panels;
end;

procedure TForm_Image.CMMouseEnter(var msg: TMessage);
begin
  Main_Form.Panel_BP.Caption	:= 'BP=' + InttoStr(BlackPoint);
  Main_Form.Panel_WP.Caption	:= 'WP=' + InttoStr(WhitePoint);
  if AAVSO_Written then
    Main_Form.Panel_Report_Written.Caption  := 'Report Written'
  else  
    Main_Form.Panel_Report_Written.Caption  := '';
  Main_Form.Panel_Ref_Background.Caption  := format('%05.1f ADUs',[Ref_Background]);  
  Main_Form.Panel_Star_Background.Caption := format('%05.1f ADUs',[Star_Background]);  
end;

procedure TForm_Image.CMMouseLeave(var msg: TMessage);
begin
  Main_Form.Clear_Bottom_Panels;
end;

// The object of this programme, to perform differential
// photometry at the specified coordinates:
// this version with snap to centroid option:
//
function TForm_Image.Measure_Flux(X,Y: Double; DoSnap: Boolean): Boolean;
var
  radius          : Integer;  // from Main_Form
  annulus         : Integer;  // from Main_Form
  gap             : Integer;  // from Main_Form
  intx,inty       : Integer;  // nearest pixel to centre of area
  xmin,ymin       : Word;     // square potential search area
  xmax,ymax       : Word;     // ditto
  pixel           : Real;     // ADU value from Image_Array[]
  Npap_inc        : Real;     // to allow for fractional pixel area
  rsqd            : Real;     // radius squared in sky annulus
  rmin            : Real;     // radius touching bounding box 
  ratio           : Real;     // proportion of pixel inside star aperture
  pixelint        : Word;     // ADU value from Image_Array[]
  Median_Pixel    : Word;
  Sum_Squared     : Real;     // for calculating SD
  SD              : Real;     // Standard Deviation of flux
  Sum1            : Real;     // to find Median_Pixel
  low_limit       : Integer;  // ADU value to eliminate > 3SD outliers
  high_limit      : Integer;  // ditto
  i,j,k           : Integer;
  rsquared1       : Real;     // annulus inner edge
  rsquared2       : Real;     // annulus outer edge
  Sky_Histogram   : Array[0..65535] of Word;  // in annulus
  Previous_Npan   : Integer;  // used to eliminate outliers
  Temp            : Real;     // to prevent integer overflow
  xsum,ysum       : Real;     // for snap to centroid
  Annulus_Flux    : Real;     // Total flux in annulus
  Aperture_Flux   : Real;     // star + sky in star aperture
begin
  k := 0;
  intx  := round(X); inty := round(Y);
  repeat
    // First check if the object is in bounds:
    //
    if not((intx>0) and (intx<XSize) and (inty>0) and (inty<YSize)) then
    begin
      Measure_Flux  := False;
      Exit;
    end;  
  
    // Truncate the aperture search area if near an edge:
    //
    radius    := Main_Form.Radius;
    if intx < radius + 1 then
      xmin  := 0
    else
      xmin  := intx - radius - 1;
    if inty < radius + 1 then
      ymin  := 0
    else
      ymin  := inty - radius - 1;
    xmax  := intx + radius + 1;
    if xmax > XSize-1 then
      xmax  := XSize-1;
    ymax  := inty + radius + 1;
    if ymax > YSize-1 then
      ymax  := YSize-1;
  
    // Then measure the total (star + sky) flux inside the star aperture:
    //
    Aperture_Flux  := 0.0;
    Npap_real   := 0.0;     // n pixels in star aperture
    xsum  := 0.0; ysum  := 0.0;
    // it is easier to scan a square of area 4*radius^2:
    // so the included circle is pi/4 of this:
    //
    for j := ymin to ymax do
      for i := xmin to xmax do
      begin
        rmin      := sqrt((i-X)*(i-X) + (j-Y)*(j-Y)) - 0.707;
        ratio     := (radius - rmin) / 1.414; // approximation
        if ratio > 0.0 then
        begin
          if ratio < 1.0 then
          begin
            pixel     := ratio * Image_Array[j,i];
            Npap_inc  := ratio
          end else
          begin
            pixel     := Image_Array[j,i];
            Npap_inc  := 1.0;
          end;
          Aperture_Flux	:= Aperture_Flux + pixel;
          xsum        := xsum + pixel*(i-X);   // find C of G
          ysum        := ysum + pixel*(j-Y);   // ditto
          Npap_real   := Npap_real + Npap_inc;
        end;
      end;

    // Move the centroid if necessary:
    //
    if DoSnap then
    begin  
      X  := X + xsum/Aperture_Flux;
      Y  := Y + ysum/Aperture_Flux;
      intx  := round(X); inty := round(Y);
      inc(k,1);
    end;
  until (k = 50) or not DoSnap;   // snap is rather slow to converge
  NewX  := X; NewY  := Y;

  // Truncate the annulus search area if near an edge:
  //
  annulus := Main_Form.Annulus;
  gap     := Main_Form.Gap;
  radius  := radius + gap;
  rsquared1 := (radius+0.5)*(radius+0.5); // inner edge of annulus
  radius  := radius + annulus;
  rsquared2 := (radius+0.5)*(radius+0.5); // outer edge of annulus
  if intx < radius then
    xmin  := 0
  else
    xmin  := intx - radius;
  if inty < radius then
    ymin  := 0
  else
    ymin  := inty - radius;
  xmax  := intx + radius;
  if xmax > XSize-1 then
    xmax  := XSize-1;
  ymax  := inty + radius;
  if ymax > YSize-1 then
    ymax  := YSize-1;

  // Now measure the background sky flux in the annulus:
  // First construct a frequency distribution to allow SD filtering:
  //
  for i := 0 to 65535 do
    Sky_Histogram[i]  := 0;
  for j := ymin to ymax do
    for i := xmin to xmax do
    begin
      rsqd  := (i-intx)*(i-intx) + (j-inty)*(j-inty);
      if (rsqd >= rsquared1) and (rsqd <= rsquared2) then
      begin  
        pixelint	:= Image_Array[j,i];
        inc(Sky_Histogram[pixelint],1);
      end;  
    end;  

  // Iterative SD filter procedure to eliminate ADU outliers
  // in the annulus by contracting low_limit and high_limit ADU values
  // until they include only values within +- 3SDs of the median value:
  //
  low_limit   := 0;
  high_limit  := 65535;
  k           := 0;
  Npan        := 99999;
  repeat  
    Previous_Npan := Npan;
    Npan          := 0;     // number of pixels in the sky annulus
    Annulus_Flux   := 0;
    Sum_Squared   := 0;
    for i:= low_limit to high_limit do    // scan all ADU values
    begin
      j           := Sky_Histogram[i];    //  number of pixels with this ADU value
      if j > 0 then
      begin  
        Npan        := Npan + j;
        Annulus_Flux := Annulus_Flux + i*j;   // Total ADU flux
        Temp        := i;                
        Temp        := Temp*i*j;            // to prevent integer overflow
        Sum_Squared := Sum_Squared + Temp;
      end;  
    end;  

    // Find the standard deviation and median:
    //
    Sky_Fluxpp  := Annulus_Flux/Npan;       // mean sky flux per pixel
    SD    := sqrt(Sum_Squared/Npan - Sky_Fluxpp*Sky_Fluxpp);
    Median_Pixel  := 0;
    Sum1          := 0;
    repeat
      Sum1 := Sum1 + Sky_Histogram[Median_Pixel];
      inc(Median_Pixel,1);
    until Sum1 >= Npan div 2;
  
    // Filter out any pixels that are > 3.0 SDs from the median:
    //
    low_limit   := Median_Pixel - round(3.0*SD);
    if low_limit < 0 then
      low_limit := 0;
    high_limit  := Median_Pixel + round(3.0*SD);
    if high_limit > 65535 then
      high_limit  := 65535;
    inc(k,1);
  until (k = 10) or (Npan = Previous_Npan);    
  Net_Flux      := Aperture_Flux - Sky_Fluxpp*Npap_real;  // subtract background
  if Net_Flux < 50 then
    Net_Flux  := 50;
  Measure_Flux  := True;  
end;  

function TForm_Image.Measure_Magnitude(): Boolean;
var
  Nstar       : Real; // net photons from star
  Nskypp      : Real; // photons per pixel of sky
  Nref        : Real; // net photons from reference star
  ObjectNoise : Real;
  RefNoise    : Real;
  ReadoutNoise: Real;
  GGain       : Real;
  Success     : Boolean;
  Ndc         : Real; // dark current e/pixel/second
  SNR_Object  : Real;
  SNR_Ref     : Real;
  Max_SD      : Real;
  Min_Mag     : Real;
  i           : Integer;
  Prev_NStar  : Real;
  Npap_int    : Integer;
  Denominator : Real;
begin
  Success := True;
  if Measure_Flux(ObjectX,ObjectY,False) then
  begin  
    GGain         := Main_Form.Gain;
    ReadoutNoise  := Main_Form.Readout_Noise;
    Ndc           := Main_Form.Dark_Current;

    Nstar         := Net_Flux * GGain;    // after background subtraction
    if Nstar  < 50 then
      Nstar := 50;
    Nskypp        := Sky_Fluxpp*GGain;
    Ref_Background  := Sky_Fluxpp;      // for display
    Npap_int      := round(Npap_real);
    ObjectNoise   := sqrt(Nstar + Npap_int*(1+Npap_int/Npan)*(Nskypp + Ndc*Exposure
                      + ReadoutNoise*ReadoutNoise + 0.289*GGain*GGain));

    if Measure_Flux(ReferenceX,ReferenceY,Main_Form.Snap_to_Centre) then
    begin  
      ReferenceX  := NewX; ReferenceY := NewY;
      Nref        := Net_Flux * GGain;    // after background subtraction
      CMag        := -2.5118864*log10(Nref/Exposure); // for AAVSO Report
      if Nref  < 50 then
        Nref := 50;
      Nskypp      := Sky_Fluxpp*GGain;
      Star_Background  := Sky_Fluxpp;      // for display
      RefNoise    := sqrt(Nref + Npap_int*(1+Npap_int/Npan)*(Nskypp + Ndc*Exposure
                      + ReadoutNoise*ReadoutNoise + 0.289*GGain*GGain));
      Magnitude   := Reference_Mag - 2.5118864* log10(Nstar/Nref);
      SNR_Object  := Nstar/ObjectNoise;
      SNR_Ref     := Nref/RefNoise;
      Std_Error   := 1.0857*sqrt(1.0/SNR_Object/SNR_Object + 1.0/SNR_Ref/SNR_ref);

      // Calculate the minimum detectable magnitude:
      //
      Max_SD      := Main_Form.Max_Error;

      // Iterative solution required:
      //
      Nstar       := 500.0;   // first guess
      i           := 0;
      repeat  
        Prev_Nstar  := Nstar;
        ObjectNoise := sqrt(Nstar + Npap_int*(1+Npap_int/Npan)*(Nskypp + Ndc*Exposure
                      + ReadoutNoise*ReadoutNoise + 0.289*GGain*GGain));
        Denominator := 0.84836*Max_SD*Max_SD - 1/SNR_Ref/SNR_Ref;
        if Denominator < 0.0001 then
        begin  
          Nstar       := 500;
          Success     := False;
          Last_Error  := ERR_REF_FLUX_NOT_MEASURED;
        end else  
          Nstar       := ObjectNoise*sqrt(1/Denominator);
        inc(i,1);
      until ((abs(Nstar-Prev_Nstar) < 0.001) or (i = 10));

      Min_Mag     := Reference_Mag - 2.5118864* log10(Nstar/Nref);
      if Magnitude > Min_Mag then
      begin
        Magnitude := Min_Mag;
        Less_Than := 1
      end else
        Less_Than := 0;
    end else begin
      Success           := False;
      Last_Error        := ERR_REF_FLUX_NOT_MEASURED;
    end;  
  end else
  begin  
    Success             := False;
    Last_Error          := ERR_STAR_FLUX_NOT_MEASURED;
  end;  
  Paint_Object();     // now its position may have been refined
  Paint_Reference();  // ditto
  if Success then
  begin  
    Last_Error          := ERR_SUCCESS;
    Measure_Magnitude   := True;
  end else 
  begin  
    Measure_Magnitude   := False;
  end;  
end;

end.
