//----------------------------------------------------------------------------
//
// Module: unit_wcs.pas (Part of Photometry Express)
//
//    Class to transform between sky positions and positions on CCD images.
//    Copied and translated from C programmes in WCSTools library written
//    by Doug Mink, SAO
//
// Author: J C Moore
//
// Copyright (C) 2013  J C Moore
//
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation, either version 3 of
// the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty
// of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
// See the GNU General Public License for more details.
//
// see http://www.gnu.org/licenses/
//
// Update History:
//    29 Aug 2011: last update by J C Moore
//    17 Nov 2013: Code commented by P Campbell-Burns
//
// Notes:
//    FITs file uses array of big endian signed 16 bit integers
//
//----------------------------------------------------------------------------{ Photometry Express

unit Unit_WCS;

interface
uses
  SysUtils,Unit_Main,Dialogs,Math;

const
  PI      = 3.14159265358979323846;
  RADDEG  = 180.0/pi;		  // degrees per radian  
  DEGRAD  = pi/180.0;     // radians per degree
  TNX_CHEBYSHEV   = 1;    // permitted types of surfaces
  TNX_LEGENDRE    = 2;
  TNX_POLYNOMIAL  = 3;
  TNX_XNONE       = 0;    // no x-terms
  TNX_XFULL       = 1;    // full x-terms
  TNX_XHALF       = 2;    // half x-terms
  BADCVAL         = 0.0;
  MAX_NITER       = 500;
  
type
  TArrayofDouble  = Array of Double;
  TPADble         = ^TArrayofDouble;

  TIRAFsurface  = Record
    valid     : Boolean;  // to overcome nil pointer problem
    xrange    : Double;   // 2. / (xmax - xmin), polynomials
    xmaxmin   : Double;   // - (xmax + xmin) / 2., polynomials
    yrange    : Double;   // 2. / (ymax - ymin), polynomials
    ymaxmin   : Double;   // - (ymax + ymin) / 2., polynomials
    stype     : Integer;  // type of curve to be fitted
    xorder    : Integer;  // order of the fit in x
    yorder    : Integer;  // order of the fit in y
    xterms    : Integer;  // cross terms for polynomials
    ncoeff    : Integer;  // total number of coefficients
    coeff     : TPADble;  // pointer to coefficient vector
    xbasis    : TPADble;  // pointer to basis functions (all x)
    ybasis    : TPADble;  // pointer to basis functions (all y)
  end;  

  TWCSRecord = Record
    coorflip  : Boolean;  // False
    crval     : Array[0..9] of double;  // CRVALn keywords
    crpix     : Array[0..9] of double;  // CRPIXn keywords
    cdelt     : Array[0..9] of double;  // CDELTn keywords
    longpole  : double;   // longitude of North Pole in degrees
    rotmat    : Boolean;  // FALSE if CDELT, CROTA; TRUE if CD
    cd        : Array[0..3] of double;  // scaling and rotation matrix
                                        // determined by plate solver
    dc        : Array[0..3] of double;  // inverse matrix
    rot       : Double;   // rotation (deg) around axis, N through E
    lngcor    : ^TIRAFsurface; // RA/longitude correction structure
    latcor    : ^TIRAFsurface; // Dec/latitude correction structure
    rodeg     : Double;   // radius of the projectiongeneating sphere
    xrefpix   : Double;
    yrefpix   : Double;
    xinc      : Double;   // X coordinate increment, degrees
    yinc      : Double;   // Y coordinate increment, degrees
  end;
  TPWCS = ^TWCSRecord;

  TWCS = class(TOBject)
  private

    function wf_gsb1leg(x: double; order: integer; k1: double; k2: double): TPADble;
    function wf_gsb1pol(x: double; order: integer): TPADble;
    function wf_gsb1cheb(x: double; order: integer; k1: double;
                            k2: double): TPADble;
    function  wf_gseval(sf: TIRAFsurface; x: double; y: double): Double;
    function  wf_gsder(sf: TIRAFsurface; x: double; y: double; nxd: integer; nyd: integer): Double;
    function  fmod(N: double; D: double): Double;

  public
    WCSRecord : TWCSRecord;
    function  tnxpos(xpix: double; ypix: double; wcs: TPWCS): TSkyPosition;
    function  tnxpix(xpos: double; ypos: double; wcs: TPWCS): TCCDPosition;

  published
    constructor Create();
  end;  

implementation

constructor TWCS.Create();
begin
  inherited Create();
end;  

// Floating Point remainder function:
//
function  TWCS.fmod(N: double; D: double): Double;
var
  Q : Integer;
begin
  Q     := Floor(N/D);
  fmod  := N - Q*D;
end;  

// Procedure to evaluate all the non-zero checbyshev function
// coefficients for a given x and order:
//
function TWCS.wf_gsb1cheb(x: double; order: integer; k1: double;
                            k2: double): TPADble;
var
  basis : TPADble;
  xnorm : Double;
  i     : Integer;
begin
  basis^[0] := 1.0;
  if order = 1 then
  begin
    wf_gsb1cheb  := basis;
    exit;
  end;  
  xnorm := (x+k1)*k2;
  basis^[1] := xnorm;
  if order = 2 then
  begin
    wf_gsb1cheb  := basis;
    exit;
  end;  
  i := 2;
  repeat
    basis^[i] := 2.0*xnorm*basis^[i-1] - basis^[i-2];
    inc(i,1);
  until i >= order;
  wf_gsb1cheb  := basis;
end;  

// Evaluate all the non-zero legendre function
// coefficients for a given x and order:
//
function TWCS.wf_gsb1leg(x: double; order: integer; k1: double; k2: double): TPADble;
var
  basis : TPADble;
  ri    : Double;
  xnorm : Double;
  i     : Integer;
begin
  basis^[0] := 1.0;
  if order = 1 then
  begin
    wf_gsb1leg  := basis;
    exit;
  end;  
  xnorm := (x+k1)*k2;
  basis^[1] := xnorm;
  if order = 2 then
  begin
    wf_gsb1leg  := basis;
    exit;
  end;  
  i := 2;
  repeat
    ri  := i;
    basis^[i] := ((2.0*ri - 1.0)*xnorm*basis^[i-1] -
                  (ri - 1.0)*basis^[i-2])/ri;
    inc(i,1);
  until i >= order;
  wf_gsb1leg  := basis;
end;  

// Evaluate all the non-zero polynomial functions
// for a single point and given order:
//
function TWCS.wf_gsb1pol(x: double; order: integer): TPADble;
var
  i     : integer;
  basis : TPADble;
begin
  basis^[0] := 1.0;
  if order = 1 then
  begin
    wf_gsb1pol  := basis;
    exit;
  end;  
  basis^[1] := x;
  if order = 2 then
  begin
    wf_gsb1pol  := basis;
    exit;
  end;  
  i := 2;
  repeat
    basis^[i] := x * basis^[i-1];
    inc(i,1);
  until i >= order;
  wf_gsb1pol  := basis;
end;  

// Evaluate the fitted surface at a single point.
// The wf.ncoeff coefficients are stored in the vector pointed
// to by sf.coeff:
//
function  TWCS.wf_gseval(sf: TIRAFsurface; x: double; y:double): Double;
var
  sum,accum       : double;
  i,ii,k          : integer;
  maxorder,xorder : integer;
begin
  // calculate the basis functions:
  //
  case sf.stype of
    TNX_CHEBYSHEV:
    begin
      sf.xbasis := wf_gsb1cheb(x, sf.xorder, sf.xmaxmin, sf.xrange);
      sf.ybasis := wf_gsb1cheb(y, sf.yorder, sf.ymaxmin, sf.yrange);
    end;  

    TNX_LEGENDRE:
    begin
      sf.xbasis := wf_gsb1leg(x, sf.xorder, sf.xmaxmin, sf.xrange);
      sf.ybasis := wf_gsb1leg(y, sf.yorder, sf.ymaxmin, sf.yrange);
    end;  

    TNX_POLYNOMIAL:
    begin
      sf.xbasis := wf_gsb1pol(x, sf.xorder);
      sf.ybasis := wf_gsb1pol(y, sf.yorder);
    end;  
  end;  

  // Initialise accumulator basis functions:
  //
  sum := 0.0;

  // Loop over the y basis functions:
  //
  if sf.xorder > sf.yorder then
    maxorder  := sf.xorder + 1
  else
    maxorder  := sf.yorder + 1;
  xorder  := sf.xorder;
  ii  := 0;
  i   := 0;
  repeat
    // Loop over the x basis functions:
    //
    accum := 0.0;
    k     := 0;
    repeat
      accum := accum + sf.coeff^[ii] * sf.xbasis^[k];
      inc(ii,1);
      inc(k,1);
    until k >= xorder;  
    accum := accum *sf.ybasis^[i];
    sum   := sum + accum;

    // elements of the coefficient vector where neither k=1 or i=1
    // are not calculated if sf.xterms = no:
    //
    if sf.xterms = TNX_XNONE then
      xorder := 1
    else if sf.xterms = TNX_XHALF then
    begin
      if(i+1+sf.xorder+1) > maxorder then
        xorder  := xorder -1;
    end;  
    inc(i,1);
  until i >= sf.yorder;  
  wf_gseval := sum;
end;

// Calculate a new surface which is a derivate of the input surface:
//
function  TWCS.wf_gsder(sf: TIRAFsurface; x: double; y: double; nxd: integer; nyd: integer): Double;
begin
  ShowMessage('wf_gsder not implemented!');
  wf_gsder  := 0.0;
end;  

// Forward transform (physical to world) gnomonic projection:
//
function  TWCS.tnxpos(xpix: double; ypix: double; wcs: TPWCS): TSkyPosition;
var
  ira             : integer;
  idec            : integer;
  x,y,r           : double;
  phi,theta       : double;
  costhe,sinthe   : double;
  dphi            : double;
  cosphi,sinphi   : double;
  dlng,z          : double;
  colatp          : double;
  coslatp,sinlatp : double;
  longp           : double;
  xs,ys,ra,dec    : double;
  xp,yp           : double;
  cosr,sinr       : double;   // temps

begin
  // convert from pixels to image coordinates:
  //
  xpix  := xpix - wcs.crpix[0];
  ypix  := ypix - wcs.crpix[1];

  if wcs.rotmat then
  begin  
    // Scale and rotate using CD matrix:
    //
    x   := xpix*wcs.cd[0] + ypix*wcs.cd[1];   // degrees
    y   := xpix*wcs.cd[2] + ypix*wcs.cd[3];   // degrees
  end else begin
    // Check axis increments - bail out if either 0:
    //
    if (wcs.cdelt[0] = 0.0) or (wcs.cdelt[1] = 0) then
    begin
      tnxpos.RA   := 0.0;
      tnxpos.Dec  := 0.0;
      exit;
    end;  

    // Scale using CDELT:
    //
    xs    := xpix*wcs.cdelt[0];
    ys    := ypix*wcs.cdelt[1];
  
    // Take out rotation from CROTA:
    //
    if not(wcs.rot = 0) then
    begin
      cosr  := cos(DEGRAD*wcs.rot);
      sinr  := sin(DEGRAD*wcs.rot);
      x     := xs*cosr - ys*sinr;
      y     := xs*sinr + ys*cosr;
    end else begin
      x     := xs;
      y     := ys;
    end;  
  end;  

  // Get the axis numbers:
  //
  if wcs.coorflip then
  begin  
    ira := 1; idec:= 0;
  end else begin  
    ira := 0; idec:= 1;
  end;
  colatp  := DEGRAD*(90.0 - wcs.crval[idec]); // radians
  coslatp := cos(colatp);
  sinlatp := sin(colatp);
  longp   := DEGRAD*wcs.longpole;             // radians

  // Compute native spherical coordinates phi and theta in radians from the
  // projected cordinates. This is the projection part of the computation:
  //
  if not(wcs.lngcor = nil) then
    xp  := x + wf_gseval(wcs.lngcor^, x, y)
  else
    xp  := x;
  if not(wcs.latcor = nil) then
    yp  := y + wf_gseval(wcs.latcor^, x, y)
  else
    yp  := y;
  x := xp;                // degrees
  y := yp;                // degrees
  r := sqrt(x*x + y*y);   // degrees

  // Compute phi:
  //
  if(r = 0) then
    phi := 0.0
  else
    phi := arctan2(x, -y);        // radians

  // Compute theta:
  //
  theta := arctan2(wcs.rodeg, r); // radians

  // Compute the celestial coordinates ra and dec from the native
  // coordinates phi and theta. This is the spherical geometry
  // part of the computation:
  //
  costhe  := cos(theta);
  sinthe  := sin(theta);
  dphi    := phi - longp;
  cosphi  := cos(dphi);
  sinphi  := sin(dphi);

  // Compute the ra:
  //
  x := sinthe * sinlatp - costhe * coslatp * cosphi;
//  if abs(x) < SPHTOL then
  if abs(x) < 0.00001 then
    x := -cos(theta+colatp) + costhe*coslatp*(1.0-cosphi);
  y := -costhe * sinphi;  
  if not(x = 0.0) or not(y = 0.0) then
    dlng  := arctan2(y,x)
  else
    dlng  := dphi + PI;
  ra  := wcs.crval[ira] + RADDEG*dlng;

  // Normalise ra:
  //
  if wcs.crval[ira] >= 0.0 then
  begin
    if ra < 0.0 then
      ra := ra + 360.0;
  end else begin
    if ra > 0.0 then
      ra := ra - 360.0;
  end;
  if ra > 360.0 then
    ra  := ra - 360.0
  else if ra < -360.0 then
    ra  := ra + 360.0;

  // Compute the dec:
  //
  if fmod(dphi, PI) = 0.0 then
  begin  
    dec := RADDEG*(theta + cosphi*colatp);
    if dec > 90.0 then
      dec := 180.0 - dec;
    if dec < -90.0 then
      dec := -180.0 - dec;
  end else begin
    z := sinthe*coslatp + costhe*sinlatp*cosphi;
    if abs(z) > 0.99 then
    begin
      if z >= 0.0 then
        dec := RADDEG*(arccos(sqrt(x*x + y*y)))
      else
        dec := RADDEG*(-arccos(sqrt(x*x + y*y)));
    end else 
      dec := RADDEG*arcsin(z);
  end;  

  tnxpos.RA := ra;
  tnxpos.Dec:= dec;
end;

// Inverse transform (world to physical) gnomonic projection:
//
function  TWCS.tnxpix(xpos: double; ypos: double; wcs: TPWCS): TCCDPosition;
var
  ira,idec,niter  : integer;
  ra,dec          : double;
  cosdec,sindec   : double;
  cosra,sinra     : double;
  x,y,phi,theta   : double;
  s,r,dphi,z,dpi  : double;
  dhalfpi,twopi,tx: double;
  xm,ym,f,fx,fy   : double;
  g,gx,gy         : double;
  denom,dx,dy     : double;
  colatp          : double;
  coslatp,sinlatp : double;
  longp,sphtol    : double;
  xresult,yresult : double;   // to prevent recursion
  cosr,sinr       : double;   // temps
begin
  // Get the axis numbers:
  //
  if wcs.coorflip then
  begin
    ira := 1; idec  := 0
  end else begin
    ira := 0; idec  := 1;
  end;

  // Compute the transformation from celestial coordinates ra and dec
  // to native coordinates phi and theta. This is the spherical geometry
  // part of the transformation:
  //
  ra      := DEGRAD*(xpos - wcs.crval[ira]);
  dec     := DEGRAD*ypos;
  cosra   := cos(ra);
  sinra   := sin(ra);
  cosdec  := cos(dec);
  sindec  := sin(dec);
  colatp  := DEGRAD*(90.0 - wcs.crval[idec]);
  coslatp := cos(colatp);
  sinlatp := sin(colatp);
  if wcs.longpole = 999.0 then
    longp := DEGRAD*180.0
  else
    longp   := DEGRAD*wcs.longpole;
  dpi     := PI;
  dhalfpi := dpi * 0.5;
  twopi   := PI + PI;
//  sphtol  := SPHTOL;
  sphtol  := 0.00001;

  // Compute phi:
  //
  x   := sindec*sinlatp - cosdec*coslatp*cosra;
  if abs(x) < sphtol then
    x := -cos(dec + colatp) + cosdec*coslatp*(1.0 - cosra);
  y   := -cosdec*sinra;
  if not(x = 0.0) or not(y =0.0) then
    dphi  := arctan2(y, x)
  else
    dphi  := ra - dpi;
  phi := longp + dphi;
  if phi > dpi then
    phi := phi - twopi
  else if phi < -dpi then
    phi := phi + twopi;

  // Compute theta:
  //
  if fmod(ra, dpi) = 0.0 then
  begin
    theta := dec + cosra*colatp;
    if theta > dhalfpi then
      theta := dpi - theta;
    if theta < -dhalfpi then
      theta := -dpi -theta;
  end else begin
    z := sindec*coslatp + cosdec*sinlatp*cosra;
    if abs(z) > 0.99 then
    begin
      if z >= 0.0 then
        theta := arccos(sqrt(x*x + y*y))
      else
        theta := -arccos(sqrt(x*x + y*y));
    end else
      theta := arcsin(z);
  end;    

  // Compute the transformation from native coordinates phi and theta
  // to projected coordinates x and y:
  //
  s := sin(theta);
  if s = 0.0 then
  begin
    x := BADCVAL; y := BADCVAL;
  end else
  begin
    r := wcs.rodeg * cos(theta) / s;
    if (wcs.lngcor = nil) and (wcs.latcor = nil) then
    begin
      if wcs.coorflip then
      begin
        y := r*sin(phi);
        x := -r*cos(phi);
      end else begin
        x := r*sin(phi);
        y := -r*cos(phi);
      end;
    end else
    begin
      xm  := r*sin(phi);
      ym  := -r*cos(phi);
      x   := xm;
      y   := ym;
      niter := 0;
      while(niter < MAX_NITER) do
      begin
        if not(wcs.lngcor = nil) then
        begin  
          f   := x + wf_gseval(wcs.lngcor^, x, y) - xm;
          fx  := wf_gsder(wcs.lngcor^, x, y, 1, 0);
          fx  := 1.0 + fx;
          fy  := wf_gsder(wcs.lngcor^, x, y, 0, 1);
        end else
        begin
          f   := x - xm;
          fx  := 1.0;
          fy  := 1.0;
        end;  
        if not(wcs.latcor = nil) then
        begin
          g   := y + wf_gseval(wcs.latcor^, x, y) - ym;
          gx  := wf_gsder(wcs.latcor^, x, y, 1, 0);
          gy  := wf_gsder(wcs.latcor^, x, y, 0, 1);
          gy  := 1.0 + gy;
        end else
        begin
          g   := y - ym;
          gx  := 0.0;
          gy  := 1.0;
        end;  

        denom := fx*gy - fy*gx;
        if denom = 0.0 then
          exit;
        dx  := (-f*gy + g*fy)/denom;
        dy  := (-g*fx + f*gx)/denom;
        x   := x + dx;
        y   := y + dy;
        if max(max(abs(dx),abs(dy)), max(abs(f),abs(y))) < 2.8e-08 then
          break;
        inc(niter,1);
      end;  
      // Reverse x and y if axes flipped:
      //
      if wcs.coorflip then
      begin
        tx  := x;
        x   := y;
        y   := tx;
      end;  
    end;  
  end;

  // Scale and rotate using inverted CD matrix:
  //
  if wcs.rotmat then
  begin
    xresult  := x*wcs.dc[0] + y*wcs.dc[1];
    yresult  := x*wcs.dc[2] + y*wcs.dc[3];
  end else
  begin
    // Correct for rotation:
    //
    if not(wcs.rot = 0.0) then
    begin
      cosr  := cos(DEGRAD*wcs.rot);
      sinr  := sin(DEGRAD*wcs.rot);
      xresult  := x*cosr + y*sinr;
      yresult  := y*cosr - x*sinr;
    end else
    begin
      xresult  := x;
      yresult  := y;
    end;  

    // Scale using CDELT:
    //
    if not(wcs.xinc = 0.0) then
      xresult  := xresult/wcs.xinc;
    if not(wcs.yinc = 0.0) then
      yresult  := yresult/wcs.yinc;
  end;    

  // Convert to pixels:
  //
  tnxpix.x  := xresult + wcs.xrefpix;
  tnxpix.y  := yresult + wcs.yrefpix;
end;

end.
